/*
    data.h - Part of psyconv, a PSION 5 data formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>       

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This file contains the declarations of all types that are used to
   represent the Word file. Variables of these types are written by the
   parse routines, and read by the generation routines. */

#ifndef PSYCONV_DATA_H
#define PSYCONV_DATA_H

#include "general.h"
#include "list.h"

typedef struct flagged_string
{
  char *data;
  u8 flags;
} *flagged_string;

typedef float cm_length_t;    /* For offsets in centimeters */

typedef float point_size_t;   /* For sizes in points */

typedef enum bool
{
  bool_false,
  bool_true, 
} bool_t;

typedef enum super_sub
{ normalscript,
  superscript,
  subscript
} super_sub_t;

typedef enum justify
{ justify_left, 
  justify_centre, 
  justify_right, 
  justify_full 
} justify_t;

typedef struct color
{
  u8 red;
  u8 green;
  u8 blue;
} color_t;

typedef struct bullet
{
  bool_t on;
  bool_t indent;
  point_size_t font_size;
  int font;
  u8 character;
  color_t color;
} bullet_t;

typedef enum border_kind
{ border_none,
  border_solid,
  border_double,
  border_dotted,
  border_striped,
  border_dotstripe,
  border_dotdotstripe
} border_kind_t;

typedef struct border
{ 
  border_kind_t kind;
  u8 thickness;
  color_t color;
} border_t;

typedef enum tab_kind
{
  tab_left,
  tab_centre,
  tab_right
} tab_kind_t;

typedef struct tab
{
  cm_length_t location;
  tab_kind_t kind;
} tab_t;


typedef list tab_list; /* of tab_t */

typedef struct all_tabs
{
  tab_t normal;
  tab_list extras;
} all_tabs_t;

typedef struct character_layout_status
{
  color_t color;
  color_t back_color;
  point_size_t font_size;
  bool_t italic;
  bool_t bold;
  super_sub_t super_sub;
  bool_t underline;
  bool_t crossed_out;
  int font; /* number as in font list */
} *character_layout_status;

typedef struct paragraph_layout_status
{
  cm_length_t indent_left;
  cm_length_t indent_right;
  cm_length_t indent_first;
  justify_t justify;
  point_size_t interline;
  bool_t interline_exact;
  point_size_t top_space;
  point_size_t bottom_space;
  bool_t on_one_page;
  bool_t together_with; /* What the heck is this? */
  bool_t on_next_page;
  bool_t no_widow_protection;
  cm_length_t left_margin;
  bullet_t bullet;
  border_t left_border;
  border_t right_border;
  border_t top_border;
  border_t bottom_border;
  all_tabs_t tabs;
} *paragraph_layout_status;

typedef struct style
{
  character_layout_status character;
  paragraph_layout_status paragraph;
  u8 hotkey;
  flagged_string name;
  u32 unknown;
} *style;

typedef struct header_section
{
  u32 uid1;
  u32 uid2;
  u32 uid3;
  u32 checksum;
  u32 sectiontable_offset;
} *header_section;

typedef struct sectiontable_entry
{
  u32 id;
  u32 offset;
} *sectiontable_entry;

typedef list sectiontable_section; /* Of sectiontable_entry */

typedef struct start_section
{
  u16 data1; /* Use unknown */
  u16 data2; /* Use unknown */
  u16 data3; /* Use unknown */
  u32 cursor_position;
  u32 data4; /* Use unknown */
} *start_section;

typedef struct appl_id_section
{
  u32 id;
  flagged_string name;
} *appl_id_section;

typedef struct in_line_layout
{
  character_layout_status layout;
  int length;
} *in_line_layout;

typedef list in_line_layouts; /* of in_line_layout */

typedef struct paragraph
{
  char *text;
  character_layout_status base_character;
  paragraph_layout_status base_paragraph;
  u8 base_style;
  in_line_layouts in_lines;
} *paragraph;

typedef list text_and_layout; /* Of paragraphs */

typedef list style_list; /* Of style */

typedef struct styles_section
{
  style normal;
  style_list styles;
} *styles_section;

typedef struct font_desc
{
  char *name;
  u8 id;
} *font_desc;

typedef list font_list; /* of font_desc */

typedef struct word_file {
  font_list fonts;
  text_and_layout paragraphs;
  start_section start_sec;
  styles_section styles_sec;
} *word_file;

#define ID_PSION_5 0x10000037
#define ID_DATA_FILE 0x1000006D
#define ID_WORD_FILE 0x1000007F

#define ID_START_SECTION 0x10000243
#define ID_APPL_ID_SECTION 0x10000089
#define ID_TEXT_SECTION 0x10000106
#define ID_LAYOUT_SECTION 0x10000143
#define ID_STYLES_SECTION 0x10000104

/* Convert to cm_length_t */
extern cm_length_t to_cm_length(u32 len);

/* Convert to point_size_t */
extern point_size_t to_point_size(u32 len);

/* Convert to bool_t */
extern bool_t to_bool(u8 val);


/* Return a clean layout_status. You can modify it at will */
extern character_layout_status basic_character_layout_status(void);

/* Return a clean layout_status. You can modify it at will */
extern paragraph_layout_status basic_paragraph_layout_status(void);

/* Clone a layout_status: the new copy is completely independent of the
   original one */

extern paragraph_layout_status clone_paragraph_layout_status
                                                 (paragraph_layout_status ls);

extern character_layout_status clone_character_layout_status
                                                 (character_layout_status ls);

/* Get a numbered style */
extern style get_style (styles_section ss, int nr);

extern font_desc get_font(font_list fl, int nr);


#endif /* def PSYCONV_DATA_H */
