/*
    parser.c - Part of psyconv, a PSION 5 data formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Parse the PSION Word file */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "data.h"
#include "general.h"
#include "parser.h"
#include "tools.h"

static u8 read_u8(const buffer buf,u32 off);
static u16 read_u16(const buffer buf,u32 off);
static u32 read_u32(const buffer buf,u32 off);
static cm_length_t read_cm_length(const buffer buf, u32 off, int *len);
static point_size_t read_point_size(const buffer buf, u32 off, int *len);
static flagged_string read_string(const buffer buf,u32 off,int *length);
static color_t read_color_t(const buffer buf, u32 off, int *length);
static int parse_font(const buffer buf, u32 off, int *len, int *font, 
                      font_list fl);
static int parse_border(const buffer buf,u32 off,int *len,border_t *border);
static int parse_header(const buffer buf,u32 off,header_section *h);
static int parse_sectiontable(const buffer buf, u32 off,
                              sectiontable_section *s);
static int parse_start_section(const buffer buf, u32 off,start_section *s);
static int parse_appl_id_section(const buffer buf, u32 off, appl_id_section *a);
static int parse_text_section(const buffer buf,u32 off,text_and_layout *t);
static int parse_paragraph_layout_codes(const buffer buf, u32 off, int *len, 
                                 paragraph_layout_status stat,font_list fl);
static int parse_character_layout_codes(const buffer buf, u32 off, int *len, 
                                 character_layout_status stat, font_list fl);
static int parse_styles_section(const buffer buf, u32 off, styles_section *ss,
                                font_list fl);
static int parse_layout_section(const buffer buf,u32 off,text_and_layout t,
                                font_list fl,styles_section ss);
static s64 offset_in_sectiontable(u32 id,const sectiontable_section s);

u8 read_u8(const buffer buf,u32 off)
{
  u8 *ptr;
  ptr = list_get(buf,off);
  if (!ptr)
    fatal("Trying byte read past the end of the file (%08x)",off);
  return *ptr;
}

u16 read_u16(const buffer buf,u32 off)
{
  u8 *ptr0,*ptr1;
  ptr0 = list_get(buf,off);
  ptr1 = list_get(buf,off+1);
  if (!ptr0 || !ptr1)
    fatal("Trying word read past the end of the file (%08x)",off);
  return *ptr0 + (*ptr1 << 8);
}

u32 read_u32(const buffer buf,u32 off)
{
  u8 *ptr0,*ptr1,*ptr2,*ptr3;
  ptr0 = list_get(buf,off);
  ptr1 = list_get(buf,off+1);
  ptr2 = list_get(buf,off+2);
  ptr3 = list_get(buf,off+3);
  if (!ptr0 || !ptr1 || !ptr2 || !ptr3)
    fatal("Trying long read past the end of the file (%08x)",off);
  return *ptr0 + (*ptr1 << 8) + (*ptr2 << 16) + (*ptr3 << 24);
}


cm_length_t read_cm_length(const buffer buf, u32 off, int *len)
{
  if (len)
    *len = 4;
  return to_cm_length(read_u32(buf,off));
}

point_size_t read_point_size(const buffer buf, u32 off, int *len)
{
  if (len)
    *len = 4;
  return to_point_size(read_u32(buf,off));
}

flagged_string read_string(const buffer buf,u32 off,int *length)
{
  u8 len;
  int i;
  flagged_string result;
  
  result = malloc(sizeof(*result));
  len = read_u8(buf,off);
  result->flags = len & 0x03;
  len = len >> 2;
  result->data = malloc(len + 1);
  for (i = 0; i < len; i++)
    result->data[i] = read_u8(buf,off+1+i);
  result->data[len] = 0;
  if (length)
    *length = len + 1;
  return result;
}

color_t read_color_t(const buffer buf, u32 off, int *length)
{
  color_t result;
  progress("Going to read a color triple");
  result.red = read_u8(buf,off);
  result.green = read_u8(buf,off+1);
  result.blue = read_u8(buf,off+2);
  debug("Color: red %02x, green %02x, blue %02x",result.red, result.green,
        result.blue);
  if (length)
    *length = 3;
  return result;
}

int parse_font(const buffer buf, u32 off, int *len, int *font, font_list fl)
{
  int res = 0;
  int strlength,i;
  char *str;
  font_desc el;
  u8 id;

  progress("Going to parse font");
  *font = -1;
  strlength = read_u8(buf,off);
  str = malloc(strlength);
  for (i = 0; i < strlength-1; i++)
    str[i] = read_u8(buf,off + 1 + i);
  str[strlength-1] = 0;
  id = read_u8(buf,off + strlength);
  debug("Found font `%s', id %02x",str,id);

  if (len)
    *len = strlength + 1;
  
  for (i = 0; (*font == -1) && (i < list_length(fl)); i ++) {
    el = list_get(fl,i);
    if (el->id == id)  {
      if (! strcmp(el->name,str)) {
        *font = id;
        goto FOUND;
      } else {
        warn("Two fonts with same id, but different names!");
        debug("id: %02x, in list: `%s', now: `%s'",id,el->name,str);
        res = -1;
        goto FOUND;
      }
    } else if (! strcmp(el->name,str)) {
      warn("Two fonts with same name, but different ids!");
      debug("font: %s, in list: %02x, now: %02x",str,el->id,id);
      res = -1;
      goto FOUND;
    }
  }
  
  el = malloc(sizeof(*el));
  el->name = str;
  el->id = id;
  *font = id;
  list_add(fl,el);
  free(el);
  return res;

FOUND:
  free(str);
  return res;
}


int parse_border(const buffer buf,u32 off,int *len,border_t *border)
{
  int res = 0;
  u32 temp;
  int leng,i;

  progress("Going to parse border data");
  temp = read_u8(buf,off);
  if (temp == 0x00)
    border->kind = border_none;
  else if (temp == 0x01)
    border->kind = border_solid;
  else if (temp == 0x02)
    border->kind = border_double;
  else if (temp == 0x03)
    border->kind = border_dotted;
  else if (temp == 0x04)
    border->kind = border_striped;
  else if (temp == 0x05)
    border->kind = border_dotstripe;
  else if (temp == 0x06)
    border->kind = border_dotdotstripe;
  else {
    warn("Unknown border type");
    res = -1;
  }
  debug("Type: %02x",temp);

  temp = read_u8(buf,off+1);
  if (!(border->kind == border_solid || border->kind == border_double ||
      temp == 0x01)) {
    warn("Border thickness specified for unlikely border type");
    res = -1;
  }
  border->thickness = temp;
  debug("Thickness: %02x",temp);

  for (i = 0; i < 2; i++) {
    temp = read_u8(buf,off + 2 + i);
    if (temp != 0x00) {
      warn("Unknown byte in border specification");
      debug("Byte %d: read %02x, expected %02x",i,temp,0x00);
      res = -1;
    }
  }

  border->color = read_color_t(buf,off+5,&leng);
  
  temp = read_u8(buf,off + 5 + leng);
  if (temp != 0x01) {
    warn("Unknown last byte in border specification");
    debug("Last byte: read %02x, expected %02x",i,temp,0x01);
    res = -1;
  }

  if (len)
    *len = leng + 6;
  
  return res;
}

int parse_header(const buffer buf,u32 off,header_section *h)
{
  int res=0;

  progress("Going to read the header section at %08x",off);
  *h = malloc(sizeof (**h));

  progress("Going to read UID1");
  (*h)->uid1 = read_u32(buf,off);
  if ((*h)->uid1 != ID_PSION_5) {
    debug("UID1: read %08x, expected %08x",(*h)->uid1,ID_PSION_5);
    warn("This may not be a Psion-5 document.");
    res = -1;
  }

  progress("Going to read UID2");
  (*h)->uid2 = read_u32(buf,off+0x04);
  if ((*h)->uid2 != ID_DATA_FILE) {
    debug("UID2: read %08x, expected %08x",(*h)->uid2,ID_DATA_FILE);
    warn("This may not be a document file");
    res = -1;
  }

  progress("Going to read UID3");
  (*h)->uid3 = read_u32(buf,off+0x08);
  if ((*h)->uid3 != ID_WORD_FILE) {
    debug("UID3: read %08x, expected %08x",(*h)->uid3,ID_WORD_FILE);
    warn("This may not be a word file");
    res = -1;
  }

  progress("Going to read Checksum");
  (*h)->checksum = read_u32(buf,off+0x0c);
  if ((*h)->checksum != 0x55089ffe) {
    debug("Checksum: read %08x, expected %08x",(*h)->checksum,0x55089ffe);
    warn("Checksum failed (ignore if one of the UIDs was off)");
    res = -1;
  }

  progress("Going to read Section Table Offset");
  (*h)->sectiontable_offset = read_u32(buf,off+0x10);
  debug("Section Table Offset: %08x", (*h)->sectiontable_offset);
  return res;
}

int parse_sectiontable(const buffer buf, u32 off,sectiontable_section *s)
{
  int i;
  u8 nr;
  int res=0;
  sectiontable_entry ste;

  progress("Going to read the section table section at %08x",off);
  *s = list_new(sizeof(*ste));

  progress("Going to read the section table length");
  nr = read_u8(buf,off);
  debug("Length: %08x",nr);

  progress("Going to read the section table entries");
  ste = malloc(sizeof(*ste));
  for (i = 0; i < nr / 2; i++) {
    ste->id = read_u32(buf,off + 1 + i * 8);
    debug("Entry %d: ID = %08x",i,ste->id);
    ste->offset = read_u32(buf,off + 5 + i * 8);
    debug("Entry %d: Offset = %08x",i,ste->offset);
    list_add(*s,ste);
  }

  free(ste);
  return res;
}

int parse_start_section(const buffer buf, u32 off,start_section *s)
{
  int res=0;

  progress("Going to read the start section at %08x",off);
  *s = malloc(sizeof(**s));

  progress("Going to read first word");
  (*s)->data1 = read_u16(buf,off);
  if ((*s)->data1 != 0x0002) {
    res = -1;
    debug("First word: read %04x, expected %04x",(*s)->data1,0x0002);
    warn("First word read of start section failed. Its use is unknown.");
  }

  progress("Going to read second word");
  (*s)->data2 = read_u16(buf,off + 0x02);
  if ((*s)->data2 != 0x0103) {
    res = -1;
    debug("Second word: read %04x, expected %04x",(*s)->data2,0x0103);
    warn("Second word read of start section failed. Its use is unknown.");
  }

  progress("Going to read third word");
  (*s)->data3 = read_u16(buf,off + 0x04);
  if ((*s)->data3 != 0x0001) {
    res = -1;
    debug("Third word: read %04x, expected %04x",(*s)->data3,0x0001);
    warn("Third word read of start section failed. Its use is unknown.");
  }
  
  progress("Going to read cursor position");
  (*s)->cursor_position = read_u32(buf,off + 0x06);
  debug("Cursor position: %08x",(*s)->cursor_position);

  progress("Going to read last long");
  (*s)->data4 = read_u32(buf,off + 0x0a);
  if ((*s)->data4 != 0x000003e8 && (*s)->data4 != 0x00000294) {
    res = -1;
    debug("Last long: read %08x, expected %08x or %08x",(*s)->data4,
           0x000003e8,0x00000294);
    warn("Last long read of start section failed. Its use is unknown.");
  } else 
    debug("Long: %08x (meaning unknown)",(*s)->data4);

  return res;
}


int parse_appl_id_section(const buffer buf, u32 off, appl_id_section *a)
{
  int res=0;

  progress("Going to read the application id section");
  *a = malloc(sizeof(**a));

  progress("Going to read the type identifier");
  (*a)->id = read_u32(buf,off);
  if ((*a)->id != ID_WORD_FILE) {
    debug("Type: read %08x, expected %08x",(*a)->id,ID_WORD_FILE);
    warn("Type field in application id section indicates a non-word document.");
  }

  progress("Going to read the application id string");
  (*a)->name = read_string(buf,off + 0x04,NULL);

  debug("Name: %s",(*a)->name->data);
  if ((*a)->name->flags != 0x02) {
    debug("Flags: read %02x, expected %02x",(*a)->name->flags,0x02);
    warn("Application ID: string marked with unknown flags");
  }

  return res;
}

int parse_text_section(const buffer buf,u32 off,text_and_layout *t)
{
  u32 len;
  paragraph para;

  int nr;
  int res = 0;
  int i,j,start;
 
  progress("Going to parse the text section at %08x",off);
  progress("Reading the text length");
  len = read_u8(buf,off);
  if ((len & 0x03) == 3) {
    len = read_u32(buf,off) >> 3;
    off += 4;
  } else if ((len & 0x03) == 1) {
    len = read_u16(buf,off) >> 2;
    off += 2;
  } else {
    len = len >> 1;
    off += 1;
  }
  debug("Length: %08x",len);

  *t = list_new(sizeof(*para));
  para = malloc(sizeof(*para));

  progress("Going to read all paragraph text");
  nr = 0;
  start = 0;
  for (i = 0; i < len; i++)
    if (read_u8(buf,off+i) == 0x06) {
      para->text = malloc(i - start + 1);
      for (j = 0; j < i - start; j++)
        para->text[j] = read_u8(buf,off + start + j);
      para->text[j] = 0;
      list_add(*t,para);

      para->text = strdup(para->text);
      for(j = 0; para->text[j]; j++)
        if (para->text[j] < 32 || para->text[j] >= 127) 
          para->text[j] = '.';
      debug("Line %d: `%s'",nr,para->text);
      free(para->text);

      start = i + 1;
      nr ++;
    }
  
  if (start != len) {
    res = -1;
    warn("Last line does not end on EOL (%d characters left)",len - start);
    para->text = malloc(len - start + 1);
    for (j = 0; j < len - start; j++)
      para->text[j] = read_u8(buf,off + start + j);
    para->text[len - start] = 0;
    list_add(*t,para);
    for(j = 0; para->text[j]; j++)
      if (para->text[j] < 32 || para->text[j] >= 127) 
        para->text[j] = '.';
    debug("Last line: `%s'",para->text);
    free(para->text);
  }
 
  free(para);

  return res;
}

int parse_paragraph_layout_codes(const buffer buf, u32 off, int *len, 
                                 paragraph_layout_status stat,font_list fl)
{
  int res=0;
  int nr,i,leng;
  u32 temp;
  u8 id,bul_leng;
  tab_t tab;

  progress("Going to read paragraph layout codes");
  nr = read_u32(buf,off);
  debug("Length: %08x",nr);
  for (i = 0; i < nr; i++) {
    id = read_u8(buf,off + 4 + i);
    switch(id) {
      case 0x02:
        progress("Going to read indent left");
        stat->indent_left = read_cm_length(buf,off + 5 + i,&leng);
        debug("Indent left: %6.3f",stat->indent_left);
        i += leng;
        break;
      case 0x03:
        progress("Going to read indent right");
        stat->indent_right = read_cm_length(buf,off + 5 + i,&leng);
        debug("Indent right: %6.3f",stat->indent_right);
        i += leng;
        break;
      case 0x04:
        progress("Going to read indent first");
        stat->indent_first = read_cm_length(buf,off + 5 + i, &leng);
        debug("Indent first: %6.3f",stat->indent_first);
        i += leng;
        break;
      case 0x05:
        progress("Going to read justify");
        temp = read_u8(buf,off+5+i);
        if (temp == 0x00)
          stat->justify = justify_left;
        else if (temp == 0x01)
          stat->justify = justify_centre;
        else if (temp == 0x02)
          stat->justify = justify_right;
        else if (temp == 0x03)
          stat->justify = justify_full;
        else {
          res = -1;
          warn("Unknown justify argument in paragraph layout codes list");
        }
        debug("Justify: %02x",temp);
        i ++;
        break;
      case 0x07:
        progress("Going to read interline");
        stat->interline = read_point_size(buf,off + 5 + i,&leng);
        debug("Interline: %6.3f",stat->interline);
        i += leng;
        break;
      case 0x08:
        progress("Going to read interline exact");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->interline_exact = to_bool(temp);
        else {
          warn("Unknown interline_exact argument in pararaph layout codes "
               "list");
          res = -1;
        }
        debug("Interline exact: %02x",temp);
        i ++;
        break;
      case 0x09:
        progress("Going to read top space");
        stat->top_space = read_point_size(buf,off + 5 + i,&leng);
        debug("Top space: %6.3f",stat->top_space);
        i += leng;
        break;
      case 0x0a:
        progress("Going to read bottom space");
        stat->bottom_space = read_point_size(buf,off + 5 + i,&leng);
        debug("Bottom space: %6.3f",stat->bottom_space);
        i += leng;
        break;
      case 0x0b:
        progress("Going to read on one page");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->on_one_page = to_bool(temp);
        else  {
          res = -1;
          warn("Unknown on_one_page argument in pararaph layout codes "
               "list");
        }
        debug("On one page: %02x",temp);
        i ++;
        break;
      case 0x0c:
        progress("Going to read together with");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->together_with = to_bool(temp);
        else {
          res = -1;
          warn("Unknown together_with argument in pararaph layout codes "
               "list");
        }
        debug("Together with: %02x",temp);
        i ++;
        break;
      case 0x0d:
        progress("Going to read on next page");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->on_next_page = to_bool(temp);
        else {
          warn("Unknown on_next_page argument in pararaph layout codes "
               "list");
          res = -1;
        }
        debug("On next page: %02x",temp);
        i ++;
        break;
      case 0x0e:
        progress("Going to read no widow protection");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->no_widow_protection = to_bool(temp);
        else  {
          warn("Unknown no_widow_protection argument in pararaph layout codes "
               "list");
          res = -1;
        }
        debug("No widow protection: %02x",temp);
        i ++;
        break;
      case 0x10:
        progress("Going to read left margin");
        stat->left_margin = read_cm_length(buf,off + 5 + i,&leng);
        debug("Left margin: %6.3f",stat->left_margin);
        i += leng;
        break;
      case 0x11:
        progress("Going to read border top");
        res |= parse_border(buf,off+5+i,&leng,&(stat->top_border));
        i += leng + 5;
        break;
      case 0x12:
        progress("Going to read border bottom");
        res |= parse_border(buf,off+5+i,&leng,&(stat->bottom_border));
        i += leng + 5;
        break;
      case 0x13:
        progress("Going to read border left");
        res |= parse_border(buf,off+5+i,&leng,&(stat->left_border));
        i += leng + 5;
        break;
      case 0x14:
        progress("Going to read border right");
        res |= parse_border(buf,off+5+i,&leng,&(stat->right_border));
        i += leng + 5;
        break;
      case 0x15:
        progress("Going to read bullet");
        stat->bullet.on = bool_true;
        bul_leng = read_u8(buf,off+5+i);
        debug("Bullet data length: %02x",bul_leng);
        stat->bullet.font_size = read_point_size(buf,off+6+i,&leng);
        debug("Font size: %6.3f",stat->bullet.font_size);
        i += leng;
        stat->bullet.character = read_u8(buf,off+6+i);
        debug("Character: %02x",stat->bullet.character);
        temp = read_u8(buf,off+7+i);
        if (temp == 0x00)
          stat->bullet.indent = bool_false;
        else if (temp == 0x01)
          stat->bullet.indent = bool_true;
        else {
          warn("Unknown intermediate bullets byte in paragraph layout codes "
               "list");
          debug("Intermediate byte: read %02x",temp);
        }
        stat->bullet.color = read_color_t(buf,off+8+i,&leng);
        i += leng;
        res |= parse_font(buf,off+8+i,&leng,&(stat->bullet.font),fl);
        i += leng + 3;
        break;
      case 0x16:
        progress("Going to read standard tabs");
        stat->tabs.normal.location = read_cm_length(buf,off+5 +i,&leng);
        i += leng;
        break;
      case 0x17:
        progress("Going to read extra tab");
        tab.location = read_cm_length(buf,off+5 +i,&leng);
        temp = read_u8(buf,off+i+5 + leng);
        if (temp == 0x00)
          tab.kind = tab_left;
        else if (temp == 0x01)
          tab.kind = tab_centre;
        else if (temp == 0x02)
          tab.kind = tab_right;
        else {
          res = -1;
          warn("Unknown kind of tab");
          debug("Tab at %06.3f of unknown type %02x",tab.location,temp);
          break;
        }
        debug("Tab at %06.3f of type %02x",tab.location,tab.kind);
        list_add(stat->tabs.extras,&tab);
        i += 1 + leng;
        break;
      default:
        warn("Unknown code in paragraph layout codes list");
        debug("Code: %02x",id);
        res = -1;
        break;
    }
  }

  if (i != nr) {
    warn("Read past end of paragraph layout codes list. I probably lost track"
         "somewhere!");
    debug("Read %d characters instead of %d",i,nr);
  }

  if (len)
    *len = nr + 4;
  return res;
}

int parse_character_layout_codes(const buffer buf, u32 off, int *len, 
                                 character_layout_status stat, font_list fl)
{
  int res=0;
  int nr,i,leng;
  u8 id;
  u32 temp;

  progress("Going to read paragraph layout codes");
  nr = read_u32(buf,off);
  debug("Length: %08x",nr);
  for (i = 0; i < nr; i++) {
    id = read_u8(buf,off + 4 + i);
    switch(id) {
      case 0x19:
        progress("Going to read color");
        stat->color = read_color_t(buf,off + 5 + i, &leng);
        i += leng;
        break;
      case 0x1a:
        progress("Going to read background color (?)");
        stat->back_color = read_color_t(buf,off + 5 + i, &leng);
        i += leng;
        break;
      case 0x1c:
        progress("Going to read font size");
        stat->font_size = read_point_size(buf,off+5+i,&leng);
        i += leng;
        debug("Size: %6.3f",stat->font_size);
        break;
      case 0x1d:
        progress("Going to read italic");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->italic = to_bool(temp);
        else {
          warn("Unknown italic argument in character layout codes "
               "list");
          res = -1;
        }
        debug("Italic: %02x",temp);
        i ++;
        break;
      case 0x1e:
        progress("Going to read bold");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->bold = to_bool(temp);
        else {
          warn("Unknown bold argument in character layout codes "
               "list");
          res = -1;
        }
        debug("Bold: %02x",temp);
        i ++;
        break;
      case 0x1f:
        progress("Going to read super_sub");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00)
          stat->super_sub = normalscript;
        else if (temp == 0x01)
          stat->super_sub = superscript;
        else if (temp == 0x02)
          stat->super_sub = subscript;
        else {
          warn("Unknown super_sub argument in character layout codes "
               "list");
          res = -1;
        }
        debug("Super_sub: %02x",temp);
        i ++;
        break;
      case 0x20:
        progress("Going to read underline");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->underline = to_bool(temp);
        else {
          warn("Unknown underline argument in character layout codes "
               "list");
          res = -1;
        }
        debug("Underline: %02x",temp);
        i ++;
        break;
      case 0x21:
        progress("Going to read crossed_out");
        temp = read_u8(buf,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->crossed_out = to_bool(temp);
        else {
          warn("Unknown crossed_out argument in character layout codes "
               "list");
          res = -1;
        }
        debug("Crossed_out: %02x",temp);
        i ++;
        break;
      case 0x22:
        progress("Going to read font");
        res |= parse_font(buf,off + 5 + i, &leng, &stat->font,fl);
        debug("Font: %02x",stat->font);
        i += leng;
        break;
      default:
        warn("Unknown code in character layout codes list");
        debug("Code: %02x",id);
        res = -1;
        break;
    }
  }

  if (i != nr) {
    warn("Read past end of character layout codes list. I probably lost track"
         "somewhere!");
    debug("Read %d characters instead of %d",i,nr);
  }

  if (len)
    *len = nr + 4;
  return res;
}

int parse_styles_section(const buffer buf, u32 off, styles_section *ss,
                         font_list fl)
{
  int res = 0;
  int len,i,nr;
  style s,this_style;
  u32 temp;
  
  progress("Going to read the styles section at %08x",off);
  (*ss) = malloc(sizeof(**ss));

  progress("Going to read style normal");
  s = malloc(sizeof(*s));
  s->name = NULL;
  s->unknown = 0x00000000;
  s->paragraph = basic_paragraph_layout_status();
  res |= parse_paragraph_layout_codes(buf,off,&len,s->paragraph,fl);
  off += len;
  s->character = basic_character_layout_status();
  res |= parse_character_layout_codes(buf,off,&len,s->character,fl);
  off += len;
  temp = read_u32(buf,off);
  if (temp & 0xffffff00) {
    warn("Normal style hotkey has unknown value");
    debug("Hotkey value %08x",temp);
    temp = 0;
    res = -1;
  } 
  s->hotkey = temp;
  if ((s->hotkey >= 32) && (s->hotkey < 127))
    debug("Hotkey: %c",s->hotkey);
  else if (s->hotkey == 0x00)
    debug("Hotkey: <none>");
  else
    debug("Hotkey: %02x",s->hotkey);
  (*ss)->normal = s;
  off += 4;

  progress("Going to read hotkeys");
  (*ss)->styles = list_new(sizeof(*s));
  s = malloc(sizeof(*s));
  nr = read_u8(buf,off);
  debug("Nummer of hotkeys: %02x",nr);
  for (i = 0; i < nr; i ++) {
    temp = read_u32(buf,off + 1 + i * 4);
    if (temp & 0xffffff00) {
      warn("Style hotkey has unknown value");
      debug("Hotkey %d value %08x",i,temp);
      temp = 0;
      res = -1;
    } 
    s->hotkey = temp;
    if ((s->hotkey >= 32) && (s->hotkey < 127))
      debug("Hotkey %d: %c",i,s->hotkey);
    else if (s->hotkey == 0x00)
      debug("Hotkey %d: <none>",i);
    else
      debug("Hotkey %d: %02x",i,s->hotkey);
    list_add((*ss)->styles,s);
  }
  free(s);
  off += 1 + nr * 4;

  progress("Going to read other styles");
  nr = read_u8(buf,off);
  if (nr != list_length((*ss)->styles)) {
    warn("Number of styles and hotkeys do not match");
    debug("%d hotkeys, %d styles",list_length((*ss)->styles),nr);
  }
  off ++;
  for (i = 0; i < nr; i++) {
    if (i >= list_length((*ss)->styles)) {
      s->hotkey = 0;
      list_add((*ss)->styles,s);
      debug("Style %d: new entry added in list",i);
    }
    this_style = list_get((*ss)->styles,i);
    this_style->name = read_string(buf,off,&len);
    debug("Style %d name: `%s'",i,this_style->name->data);
    if (this_style->name->flags != 0x02) {
      warn("Style name: string marked with unknown flags");
      debug("Style %d flags: read %02x, expected %02x",
            i,this_style->name->flags,0x02);
    }

    off += len;
    this_style->unknown = read_u32(buf,off);
    debug("Style %d unknown id: %08x",i,this_style->unknown);
    off += 4;
    temp = read_u32(buf,off);
    if (temp != 0x00000000) {
      warn("Unexpected value in style description");
      debug("Style %d long: read %08x, expected %08x",i,temp,0x00000000);
    }
    off += 4;
    this_style->character = basic_character_layout_status();
    res |= parse_character_layout_codes(buf,off,&len,this_style->character,fl);
    off += len;
    this_style->paragraph = basic_paragraph_layout_status();
    res |= parse_paragraph_layout_codes(buf,off,&len,this_style->paragraph,fl);
    off += len;
  }

  return res;
}


void no_layout_section(text_and_layout t,styles_section ss)
{
  paragraph para;
  int i;

  for (i = 0; i < list_length(t); i ++) {
    para = list_get(t,i);
    para->in_lines = list_new(sizeof(struct in_line_layout));
    para->base_style = 0;
    para->base_character = clone_character_layout_status(ss->normal->character);
    para->base_paragraph = clone_paragraph_layout_status(ss->normal->paragraph);
  }
}

/* First do a parse_text_section, or you will get into trouble here */
int parse_layout_section(const buffer buf,u32 off,text_and_layout t,
                         font_list fl,styles_section ss)
{
  int res = 0;
  int len,leng,nr,i,j,k,l;
  u32 temp;

  typedef struct anon_style
  {
    int nr;
    u8 base_style;
    character_layout_status character;
    paragraph_layout_status paragraph;
  } *anon_style;

  typedef list anon_style_list; /* of anon_style */

  anon_style_list anon_styles;
  struct anon_style this_anon;
  in_line_layout this_in_line;
  anon_style anon_ptr = NULL;
  u32 *nr_els;
  paragraph par;

  progress("Going to read the layout section at %08x",off);
  progress("Going to read the preamble");
  temp = read_u16(buf,off);
  if (temp != 0x0001) {
    res = -1;
    debug("First word: read %08x, expected %08x",temp,0x0001);
    warn("First word read of the preamble of the layout section failed. "
         "Its use is unknown.");
  }

  progress("Going to read paragraph type list");
  anon_styles = list_new(sizeof(this_anon));
  progress("Going to read paragraph type list length");
  nr = read_u8(buf,off + 0x02);
  debug("Length: %02x",nr);
  progress("Going to read paragraph type list elements");
  len = 0x03;
  for (i = 0; i < nr; i ++) {
    this_anon.nr = read_u32(buf,off + len);
    debug("Element %d Type: %08x",i,this_anon.nr);
    len += 0x04;
    this_anon.base_style = read_u8(buf, off + len +0x04 +read_u32(buf,off+len));
    debug("Element %d Base style: %02x",i,this_anon.base_style);
    this_anon.paragraph = clone_paragraph_layout_status(
                                 get_style(ss,this_anon.base_style)->paragraph);
    res |= parse_paragraph_layout_codes(buf,off+len,&leng,this_anon.paragraph,
                                        fl);
    debug("Element %d Base style: %02x",i,this_anon.base_style);
    len += leng + 1;
    this_anon.character = clone_character_layout_status(
                                 get_style(ss,this_anon.base_style)->character);
    res |= parse_character_layout_codes(buf,off+len,&leng,this_anon.character,
                                        fl);
    list_add(anon_styles,&this_anon);
    len += leng;
  }

  progress("Going to parse the paragraph element sequence");
  progress("Going to read the number of paragraphs");
  nr = read_u32(buf,off+len);
  if (nr != list_length(t)) {
    warn("Number of text paragraphs and paragraph elements does not match");
    debug("%d text paragraphs, %d paragraph elements",list_length(t),nr);
  } 

  len += 4;

  nr_els = malloc(nr * sizeof(*nr_els));
  progress("Going to read the paragraph elements");
  for (i = 0; i < nr; i ++) {
    if (i >= list_length(t)) {
      par = malloc(sizeof(*par));
      par->text = strdup("");
      list_add(t,par);
      free(par);
      res = -1;
      debug("Paragraph %d: new entry added in list",i);
    }
    par = list_get(t,i);
    
    temp = read_u32(buf,off+len);
    if (temp != strlen(par->text)+1) {
      res = -1;
      warn("Disagreement of the length of paragraph %d in layout section",i);
      debug("Paragraph %d length: layout section says %d, counted %d",i,
            temp,strlen(par->text)+1);
    } else
      debug("Paragraph %d length: %d",i,temp);
    
    temp = read_u8(buf,off + len + 0x04);
    if (temp != 0x00) {
      debug("Paragraph %d type: %02x",i,temp);
      for (j = 0; j < list_length(anon_styles); j ++) {
        anon_ptr = list_get(anon_styles,j);
        if (temp == anon_ptr->nr)
          break;
      }
      if (j == list_length(anon_styles)) {
        warn("Unknown paragraph type for paragraph %d",i);
        debug("Paragraph %d unknown type %02x - falling back on normal style",
              i,temp);
        par->base_style = 0;
        par->base_character = clone_character_layout_status
                                                       (ss->normal->character);
        par->base_paragraph = clone_paragraph_layout_status
                                                       (ss->normal->paragraph);
      } else {
        par->base_style = anon_ptr->base_style;
        par->base_character = clone_character_layout_status
                                                       (anon_ptr->character);
        par->base_paragraph = clone_paragraph_layout_status
                                                       (anon_ptr->paragraph);
      }
      nr_els[i] = 0;
      len += 5;
    } else {
      debug("Paragraph %d: not based on a paragraph type",i);
      par->base_style = read_u8(buf,off+len+0x09+read_u32(buf,off+len+0x05));
      debug("Paragraph %d Base style: %02x",i,par->base_style);
      par->base_character = clone_character_layout_status(
                                     get_style(ss,par->base_style)->character);
      par->base_paragraph = clone_paragraph_layout_status(
                                     get_style(ss,par->base_style)->paragraph);
      res |= parse_paragraph_layout_codes(buf,off + len + 0x05, &leng,
                                          par->base_paragraph,fl);
      nr_els[i] = read_u32(buf,off + len + leng + 0x06);
      debug("Paragraph %d: %d layout elements",i,nr_els[i]);
      len += leng + 0x0a;
    }
  }

  progress("Going to read the layout element sequence");
  progress("Going to read the number of elements");
  nr = read_u32(buf,off+len);
  debug("Elements: %08x",nr);
  len += 0x04;

  k = 0;
  this_in_line = malloc(sizeof(*this_in_line));
  for (i = 0; i < list_length(t); i++) {
    l = 0;
    par = list_get(t,i);
    par->in_lines = list_new(sizeof(*this_in_line));
    progress("Paragraph %d: %d elements (current: %d, left: %d)",i,nr_els[i],
             k, nr - k);
    this_in_line->layout = par->base_character;
    if (k + nr_els[i] > nr) {
      res = -1;
      warn("Paragraph %d: layout elements list exhausted, can't register "
           "internal layout");
    }
    for (j = k; (j < nr) && (j < k + nr_els[i]); j++) {
      this_in_line->layout = clone_character_layout_status
                                                       (this_in_line->layout);
      temp = read_u8(buf,off+len);
      debug("Paragraph %d entry %d type: %02x",i,j,temp);
      this_in_line->length = read_u32(buf,off+len+1);
      debug("Paragraph %d entry %d length: %08x",i,j,this_in_line->length);
      l += this_in_line->length;
      this_in_line->layout->italic = par->base_character->italic;
      this_in_line->layout->bold = par->base_character->bold;
      this_in_line->layout->super_sub = par->base_character->super_sub;
      this_in_line->layout->underline = par->base_character->underline;
      this_in_line->layout->crossed_out = par->base_character->crossed_out;
      res |= parse_character_layout_codes(buf,off + len + 5, &leng,
                                          this_in_line->layout,fl);
      if (temp == 0x01) {
        debug("Paragraph %d entry %d: skipped object data",i,j);
        len += 0x10;
      } else if (temp != 0x00) {
        debug("Paragraph %d entry %d: unknown type. I may get confused now.");
        nr = -1;
      }
      list_add(par->in_lines,this_in_line);
      len += 5 + leng;
    }
    k = j;
  }
  if (k != nr) {
    res = -1;
    warn("Some unused layout elements are left");
    debug("Elements: Paragraphs wanted %d, we found %d",nr,k);
  }
  free(this_in_line);

  return res;
}

s64 offset_in_sectiontable(u32 id,const sectiontable_section s)
{
  int i;
  sectiontable_entry ep;
  for (i = 0; i < list_length(s); i++) {
    ep = list_get(s,i),id;
    if (id == ep->id)
      return ep->offset;
  }
  return -1;
}


int parse(const buffer buf,word_file *wf)
{

  header_section h;
  sectiontable_section s;
  appl_id_section a;
  int off;

  int res = 0;
  *wf = malloc(sizeof(**wf));

  res = parse_header(buf,0,&h);
  res |= parse_sectiontable(buf,h->sectiontable_offset,&s);

  progress("Looking for the Start section");
  off = offset_in_sectiontable(ID_START_SECTION,s);
  if (off < 0) {
   warn("Start section not found in the section table");
    res = -1;
  } else {
    debug("Start section at offset %08x",off);
    res |= parse_start_section(buf,off,&((*wf)->start_sec));
  }

  progress("Looking for the Application ID section");
  off = offset_in_sectiontable(ID_APPL_ID_SECTION,s);
  if (off < 0) {
    warn("Application ID section not found in the section table");
    res = -1;
  } else {
    debug("Application ID section at offset %08x",off);
    res |= parse_appl_id_section(buf,off,&a);
  }

  (*wf)->fonts = list_new(sizeof(struct font_desc));

  progress("Looking for the Styles section");
  off = offset_in_sectiontable(ID_STYLES_SECTION,s);
  if (off < 0) {
    warn("Styles section not found in the section table");
    res = -1;
  } else {
    debug("Styles section at offset %08x",off);
    res |= parse_styles_section(buf,off,&((*wf)->styles_sec),(*wf)->fonts);
  }

  progress("Looking for the Text section");
  off = offset_in_sectiontable(ID_TEXT_SECTION,s);
  if (off < 0) {
    warn("Text section not found in the section table");
    res = -1;
  } else {
    debug("Text section at offset %08x",off);
    res |= parse_text_section(buf,off,&((*wf)->paragraphs));
  }

  progress("Looking for the (optional) Layout section");
  off = offset_in_sectiontable(ID_LAYOUT_SECTION,s);
  if (off < 0) {
    debug("Layout section not found in the section table");
    no_layout_section((*wf)->paragraphs,(*wf)->styles_sec);
  } else {
    debug("Layout section at offset %08x",off);
    res |= parse_layout_section(buf,off,(*wf)->paragraphs,(*wf)->fonts,
                                (*wf)->styles_sec);
  }
  return res;
}
